#include <vcl.h>
#include <stdlib.h>
#include <values.h>
#include "Tools.cpp"
#include "Camera.cpp"
#include "TTrajectory.cpp"

// -----------------------------   --------------------------------

class TPlanet {
public:

	double	Mass;					//    
    double	Radius;					// 
    AnsiString Name;				// 
    bool	ShowName;				//     

    double	X, Y, Z;				//  
    double	Vx, Vy, Vz;				//   
    double	NextX, NextY, NextZ;	//    . 
    double	NextVx, NextVy, NextVz;	//     . 

    TTrajectory *Trajectory;		//  
    TColor	TrajColor;				//  
    bool	HasTrajectory;			//   

    TImage	*Image;					//     
    TColor	Color;					// 
    TColor  EdgeColor;				//     
    TColor	ActColor;				//  
    bool	Selected;				// 
    bool	Active;					// 
    bool	Inside;					//    
    int     ArrowS, ArrowH;			//      

    TPlanet()
    {Trajectory=new TTrajectory; Image=NULL; Name="";};

    TPlanet(TImage*,CAMERA*,double);	//     
    TPlanet(TImage*,TColor,double,double,double,double,double,double,double,double);
    									//  
    ~TPlanet(){delete Trajectory;};
    bool Touch(int,int,CAMERA*,bool);	//     
    bool InsideRect(TRect*,CAMERA*);	//   
    void DrawHalo(CAMERA*,int,TColor);	//  
    void DrawSelection(CAMERA*);		//   
    void DrawActivation(CAMERA*);		//   
    void Draw(CAMERA*,bool,bool,Byte,Byte**);	//   
    void DrawTrajectory(CAMERA*);		//  
};

// ----------------------------------------------------------------------------
// :

//     :
TPlanet::TPlanet(TImage *image, CAMERA *Camera, double Fx)
{
	// Fx -       
    Image = image;  ArrowS = 10;  ArrowH = 2;
    Selected = false;  Active = false;  Inside = false;

	int W = (Image!=NULL) ? Image->Width  : 800;	//  
	int H = (Image!=NULL) ? Image->Height : 600;	//  
    double F = double(W)/2/tan(Camera->Phi);		//  
    double **R = Camera->R;							//  

	double Xp = W*(double(rand())/RAND_MAX - 0.5);
    double Yp = H*(double(rand())/RAND_MAX - 0.5);
    double C  = Fx*sqrt(double(rand())/RAND_MAX);
    double XR = C*Xp;
    double YR = C*Yp;
    double ZR = -C*F;
    X = R[0][0]*XR + R[1][0]*YR + R[2][0]*ZR + Camera->X0;		NextX = X;
    Y = R[0][1]*XR + R[1][1]*YR + R[2][1]*ZR + Camera->Y0;		NextY = Y;
    Z = R[0][2]*XR + R[1][2]*YR + R[2][2]*ZR + Camera->Z0;		NextZ = Z;
    Vx = random(W/5)-W/10;				NextVx = Vx;
    Vy = random(H/5)-H/10;				NextVy = Vy;
    Vz = random(int(F/5))-int(F/10);	NextVz = Vz;
    Radius = random(min(40.0,max(W,H)/20.0)) + 1;
    Mass   = Radius*Radius*Radius;
    Color  = TColor(random(256)*256*256 + random(256)*256 + random(256));
    EdgeColor = clBlack;  ActColor = clBlue;  TrajColor = Color;
    Trajectory = new TTrajectory;   HasTrajectory = false;
    Name = "";  ShowName = false;
};

//    :
TPlanet::TPlanet
(TImage *image,TColor color,double radius,double mass,double x,double y,double z,double vx,double vy,double vz)
{
    Image = image;  ArrowS = 10;  ArrowH = 2;
    Selected = false;  Active = false;  Inside = false;
	radius = (radius>0) ? radius : 0;
    mass = (mass>0) ? mass : 1;

	X  = x;		NextX  = x;		Mass   = mass;
    Y  = y;		NextY  = y;		Radius = radius;
    Z  = z;     NextZ  = z;
    Vx = vx;	NextVx = vx;	Color  = color;
    Vy = vy;	NextVy = vy;	EdgeColor = clBlack;  ActColor = clBlue;
    Vz = vz;    NextVz = vz;    TrajColor = Color;

    Trajectory = new TTrajectory;   HasTrajectory = false;
    Name = "";  ShowName = false;
};

// ----------------------------------------------------------------------------
//     :

bool TPlanet::Touch(int MX, int MY, CAMERA *Camera, bool Design)
{
	// MX, MY -    
	if (Image == NULL) return false;
	int W = Image->Width;
    int H = Image->Height;
    //   :
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double F = double(W)/2/tan(Camera->Phi);		//  
    double **R = Camera->R;							//  

    //      :
    double XR = R[0][0]*(X-X0) + R[0][1]*(Y-Y0) + R[0][2]*(Z-Z0);
    double YR = R[1][0]*(X-X0) + R[1][1]*(Y-Y0) + R[1][2]*(Z-Z0);
    double ZR = R[2][0]*(X-X0) + R[2][1]*(Y-Y0) + R[2][2]*(Z-Z0);
    double C = -ZR/F;
    double MinInt = 1.0/(MAXINT - 4092);
    if (!(C > MinInt*((XR>0)?XR:-XR) && C > MinInt*((YR>0)?YR:-YR) && C > 0)) return false;
    //      :
    double d2 = (XR*XR + YR*YR)/(C*C); double L2 = XR*XR + YR*YR + ZR*ZR;
    if (L2-Radius*Radius <= 0) return false;
    int Rad = floor(Radius*(d2+F*F)/(F*sqrt(L2-Radius*Radius)+Radius*sqrt(d2)));
    if (!(0<=Rad && Rad<=4000)) return false;
    Rad += 5*(Design*Selected) + 3*(Active*(!Selected || !Design));
    Rad = (Rad==0) ? 1 : Rad;
    //      :
    int Xp = W/2 + trunc(XR/C);
    int Yp = H/2 - trunc(YR/C);
    //    - :
    return double(Xp-MX)*(Xp-MX) + double(Yp-MY)*(Yp-MY) <= double(Rad)*Rad;
};

// ----------------------------------------------------------------------------
//     .    
// .

bool TPlanet::InsideRect(TRect *Rect, CAMERA *Camera)
{
	if (Image == NULL) return false;
	int W = Image->Width;
    int H = Image->Height;
    //   :
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double F = double(W)/2/tan(Camera->Phi);		//  
    double **R = Camera->R;							//  

    //      :
    double XR = R[0][0]*(X-X0) + R[0][1]*(Y-Y0) + R[0][2]*(Z-Z0);
    double YR = R[1][0]*(X-X0) + R[1][1]*(Y-Y0) + R[1][2]*(Z-Z0);
    double ZR = R[2][0]*(X-X0) + R[2][1]*(Y-Y0) + R[2][2]*(Z-Z0);
    double C = -ZR/F;
    double MinInt = 1.0/(MAXINT - 4092);
    if (!(C > MinInt*((XR>0)?XR:-XR) && C > MinInt*((YR>0)?YR:-YR) && C > 0)) return false;
    //      :
    double d2 = (XR*XR + YR*YR)/(C*C); double L2 = XR*XR + YR*YR + ZR*ZR;
    if (L2-Radius*Radius <= 0) return false;
    int Rad = floor(Radius*(d2+F*F)/(F*sqrt(L2-Radius*Radius)+Radius*sqrt(d2)));
    if (!(0<=Rad && Rad<=4000)) return false;
    //      :
    int Xp = W/2 + trunc(XR/C);
    int Yp = H/2 - trunc(YR/C);

	//	 :
    double Left   = Rect->Left;
    double Top    = Rect->Bottom;
    double Right  = Rect->Right;
    double Bottom = Rect->Top;

	//	      :
	if (Left - Rad <= Xp && Xp <= Right + Rad &&
        Bottom - Rad <= Yp && Yp <= Top + Rad ) {
        //	     .
        //	  :
        if (Left - Rad <= Xp && Xp <= Left && Top <= Yp && Yp <= Top + Rad )
        	return ( (Xp-Left)*(Xp-Left) + (Yp-Top)*(Yp-Top) <= Rad*Rad );
        //	  :
        if (Right <= Xp && Xp <= Right + Rad && Top <= Yp && Yp <= Top + Rad )
        	return ( (Xp-Right)*(Xp-Right) + (Yp-Top)*(Yp-Top) <= Rad*Rad );
        //	  :
        if (Right <= Xp && Xp <= Right + Rad && Bottom - Rad <= Yp && Yp <= Bottom  )
        	return ( (Xp-Right)*(Xp-Right) + (Yp-Bottom)*(Yp-Bottom) <= Rad*Rad );
        //	  :
        if (Left - Rad <= Xp && Xp <= Left && Bottom - Rad <= Yp && Yp <= Bottom  )
        	return ( (Xp-Left)*(Xp-Left) + (Yp-Bottom)*(Yp-Bottom) <= Rad*Rad );
    }
    else return false;
    return true;
};

// ----------------------------------------------------------------------------
//    HaloColor  h :

void TPlanet::DrawHalo(CAMERA *Camera, int h, TColor HaloColor)
{
	if (Image == NULL) return;
	int W = Image->Width;
    int H = Image->Height;
    //   :
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double F = double(W)/2/tan(Camera->Phi);		//  
    double **R = Camera->R;							//  

    //      :
    double XR = R[0][0]*(X-X0) + R[0][1]*(Y-Y0) + R[0][2]*(Z-Z0);
    double YR = R[1][0]*(X-X0) + R[1][1]*(Y-Y0) + R[1][2]*(Z-Z0);
    double ZR = R[2][0]*(X-X0) + R[2][1]*(Y-Y0) + R[2][2]*(Z-Z0);
    double C = -ZR/F;
    double MinInt = 1.0/(MAXINT - 4092);
    if (!(C > MinInt*((XR>0)?XR:-XR) && C > MinInt*((YR>0)?YR:-YR) && C > 0)) return;
    //      :
    double d2 = (XR*XR + YR*YR)/(C*C); double L2 = XR*XR + YR*YR + ZR*ZR;
    if (L2-Radius*Radius <= 0) return;
    int Rad = floor(Radius*(d2+F*F)/(F*sqrt(L2-Radius*Radius)+Radius*sqrt(d2)));
    if (!(0<=Rad && Rad<=4000)) return;
    //      :
    int Xp = W/2 + trunc(XR/C);
    int Yp = H/2 - trunc(YR/C);

    Image->Canvas->Pen->Mode  = pmNotXor;
    Image->Canvas->Pen->Color = HaloColor;
    Image->Canvas->Brush->Style = bsClear;
    Image->Canvas->Ellipse(Xp - (Rad+h),   Yp - (Rad+h),
                           Xp + (Rad+h)+1, Yp + (Rad+h)+1 );
};

// ----------------------------------------------------------------------------
//   :

void TPlanet::DrawSelection(CAMERA *Camera)
{
	DrawHalo(Camera,5,EdgeColor);
};

// ----------------------------------------------------------------------------
//   :

void TPlanet::DrawActivation(CAMERA *Camera)
{
	DrawHalo(Camera,3,ActColor);
};

// ----------------------------------------------------------------------------
//  :

void TPlanet::Draw(CAMERA *Camera, bool Design, bool Draw3D, Byte Brightness, Byte **BMP)
{
	if (Image == NULL) return;
	int W = Image->Width;
    int H = Image->Height;
    //   :
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double F = double(W)/2/tan(Camera->Phi);		//  
    double **R = Camera->R;							//  

    //      :
    double XR = R[0][0]*(X-X0) + R[0][1]*(Y-Y0) + R[0][2]*(Z-Z0);
    double YR = R[1][0]*(X-X0) + R[1][1]*(Y-Y0) + R[1][2]*(Z-Z0);
    double ZR = R[2][0]*(X-X0) + R[2][1]*(Y-Y0) + R[2][2]*(Z-Z0);
    double C = -ZR/F;
    double MinInt = 1.0/(MAXINT - 4092);
    if (!(C > MinInt*((XR>0)?XR:-XR) && C > MinInt*((YR>0)?YR:-YR) && C > 0)) return;
    //      :
/*
    if (ZR <= Radius) return;    double dZR = sqrt(ZR*ZR - Radius*Radius);
    if (dZR <= 2.5E-4*F*Radius) return;
    int Rad = floor(F*Radius/dZR);
*/
    double d2 = (XR*XR + YR*YR)/(C*C); double L2 = XR*XR + YR*YR + ZR*ZR;
    if (L2-Radius*Radius <= 0) return;
    int Rad = floor(Radius*(d2+F*F)/(F*sqrt(L2-Radius*Radius)+Radius*sqrt(d2)));
    if (!(0<=Rad && Rad<=4000)) return;
    //      :
    int Xp = W/2 + trunc(XR/C);
    int Yp = H/2 - trunc(YR/C);

    int Rad2 = Rad + 5*Selected + 3*(!Selected)*Active + 1;
    if (!( -Rad2 < Xp && Xp < W + Rad2 &&
           -Rad2 < Yp && Yp < H + Rad2 )) return;

    Image->Canvas->Pen->Mode    = pmCopy;
    Image->Canvas->Brush->Color = Color;

	//  :
	if (Rad>0) {
        if (Draw3D) {
        	int i,j,r;
            int Rad2 = Rad*Rad;
            int R = int(Color%256);
            int G = int((Color/256)%256);
            int B = int(((Color/256)/256)%256);
            int hR = 255 - (255 - R)/(Brightness+1);
            int hG = 255 - (255 - G)/(Brightness+1);
            int hB = 255 - (255 - B)/(Brightness+1);
		    //      :
            for (i=-Rad+1; i<Rad; i++) if (Yp+i>=0 && Yp+i<H)
            	BMP[Yp+i] = (Byte*) Image->Picture->Bitmap->ScanLine[Yp+i];
            for (i=-Rad+1; i<Rad; i++)  for (j=-Rad+1; j<Rad; j++) {
                r = i*i + j*j;
            	if (r<Rad2 && Yp+i>=0 && Yp+i<H && Xp+j>=0 && Xp+j<W) {
                	BMP[Yp+i][3*(Xp+j)]   = Byte(hB + ((B-hB)*r)/Rad2);
                	BMP[Yp+i][3*(Xp+j)+1] = Byte(hG + ((G-hG)*r)/Rad2);
                	BMP[Yp+i][3*(Xp+j)+2] = Byte(hR + ((R-hR)*r)/Rad2);
                }
            }
        }
    	Image->Canvas->Pen->Color   = EdgeColor;
        Image->Canvas->Brush->Style = (Draw3D) ? bsClear : bsSolid;
        Image->Canvas->Ellipse(Xp - Rad,   Yp - Rad,
                               Xp + Rad+1, Yp + Rad+1 );
    }
    if (Rad==0) {
    	Image->Canvas->Pen->Color = EdgeColor;
    	Image->Canvas->MoveTo(Xp,Yp);
        Image->Canvas->LineTo(Xp+1,Yp);
    }

    //   :
    if (Design) {
    	Image->Canvas->Pen->Mode = pmNotXor;
    	Image->Canvas->Pen->Color = EdgeColor;
    	double V = sqrt(Vx*Vx + Vy*Vy + Vz*Vz);
        //    :
    	double Vx0 = (V>MINDOUBLE) ? X + Radius*Vx/V : X + Radius;
        double Vy0 = (V>MINDOUBLE) ? Y + Radius*Vy/V : Y;
        double Vz0 = (V>MINDOUBLE) ? Z + Radius*Vz/V : Z;
        //    :
    	double Vx1 = (V>MINDOUBLE) ? Vx0 + Vx : X + Radius;
        double Vy1 = (V>MINDOUBLE) ? Vy0 + Vy : Y;
        double Vz1 = (V>MINDOUBLE) ? Vz0 + Vz : Z;
        //        :
	    double Vx0R = R[0][0]*(Vx0-X0) + R[0][1]*(Vy0-Y0) + R[0][2]*(Vz0-Z0);
	    double Vy0R = R[1][0]*(Vx0-X0) + R[1][1]*(Vy0-Y0) + R[1][2]*(Vz0-Z0);
	    double Vz0R = R[2][0]*(Vx0-X0) + R[2][1]*(Vy0-Y0) + R[2][2]*(Vz0-Z0);
	    C = -Vz0R/F;
        if (C <= 0) goto OREOL;
    	if (!(C > MinInt*((Vx0R>0)?Vx0R:-Vx0R) && C > MinInt*((Vy0R>0)?Vy0R:-Vy0R))) goto OREOL;
        //    :
        double Vx0p = W/2.0 + Vx0R/C;
        double Vy0p = H/2.0 - Vy0R/C;
        //        :
	    double Vx1R = R[0][0]*(Vx1-X0) + R[0][1]*(Vy1-Y0) + R[0][2]*(Vz1-Z0);
	    double Vy1R = R[1][0]*(Vx1-X0) + R[1][1]*(Vy1-Y0) + R[1][2]*(Vz1-Z0);
	    double Vz1R = R[2][0]*(Vx1-X0) + R[2][1]*(Vy1-Y0) + R[2][2]*(Vz1-Z0);
	    C = -Vz1R/F;
        //     , :
        if (C <= 0) {
        	double k = Vz0R/(Vz0R - Vz1R);
        	Vx1R = Vx0R + k*(Vx1R - Vx0R);
        	Vy1R = Vy0R + k*(Vy1R - Vy0R);
        	Vz1R = 0;  C = 0;
        }
    	if (!(C > MinInt*((Vx1R>0)?Vx1R:-Vx1R) && C > MinInt*((Vy1R>0)?Vy1R:-Vy1R))) C = 10*MinInt*max(1.0, max(((Vx1R>0)?Vx1R:-Vx1R), ((Vy1R>0)?Vy1R:-Vy1R)));
        //    :
        double Vx1p = W/2.0 + Vx1R/C;
        double Vy1p = H/2.0 - Vy1R/C;
        //	:
        int Vx0i = trunc(Vx0p);  int Vy0i = trunc(Vy0p);
        int Vx1i = trunc(Vx1p);  int Vy1i = trunc(Vy1p);
    	//	   , :
        double Vxp = Vx1p - Vx0p;
        double Vyp = Vy1p - Vy0p;
        double Vp = sqrt(Vxp*Vxp + Vyp*Vyp);
        double D = sqrt(double(W)*W + double(H)*H) + 2*ArrowH;
        if (Vp > D) {
    		Image->Canvas->MoveTo(int(Vx0p + Vxp*D/Vp), int(Vy0p + Vyp*D/Vp));
        	Image->Canvas->LineTo(Vx0i, Vy0i);
            goto OREOL;
        }
    	//	 ,   :
        if (V>MINDOUBLE && Vp>MINDOUBLE) {
        	Image->Canvas->MoveTo(Vx1i, Vy1i);
        	Image->Canvas->LineTo(Vx0i, Vy0i);
        	Image->Canvas->MoveTo(Vx1i, Vy1i);
            Image->Canvas->LineTo(trunc(Vx1p - 1/Vp*(ArrowS*Vxp + ArrowH*Vyp)),
                              	  trunc(Vy1p - 1/Vp*(ArrowS*Vyp - ArrowH*Vxp)) );
            Image->Canvas->MoveTo(Vx1i, Vy1i);
        }
        else {
        	Image->Canvas->MoveTo(Xp + Rad, Yp);
        	Image->Canvas->LineTo(Xp + Rad - ArrowS, Yp-ArrowH);
            Image->Canvas->MoveTo(Xp + Rad, Yp);
        }
        if (V>MINDOUBLE && Vp>MINDOUBLE)
             Image->Canvas->LineTo(trunc(Vx1p - 1/Vp*(ArrowS*Vxp - ArrowH*Vyp)),
                                   trunc(Vy1p - 1/Vp*(ArrowS*Vyp + ArrowH*Vxp)) );
        else Image->Canvas->LineTo(Xp + Rad - ArrowS, Yp+ArrowH);
    }
OREOL:
    //   :
    Image->Canvas->Pen->Style = psSolid;
    if (Selected && Design) {
    	Image->Canvas->Pen->Color   = EdgeColor;
        Image->Canvas->Brush->Style = bsClear;
        Image->Canvas->Ellipse(Xp - (Rad+5),   Yp - (Rad+5),
                               Xp + (Rad+5)+1, Yp + (Rad+5)+1 );
    }
	//   :
    if (Active) {
    	Image->Canvas->Pen->Mode = pmNotXor;
    	Image->Canvas->Pen->Color = ActColor;
    	Image->Canvas->Brush->Style = bsClear;
    	Image->Canvas->Ellipse(Xp - (Rad+3),   Yp - (Rad+3),
                               Xp + (Rad+3)+1, Yp + (Rad+3)+1 );
    }

    //  :
    if ((Design || ShowName) && Name != "") {
    	int Wt = Image->Canvas->TextWidth(Name);
        int Ht = Image->Canvas->TextHeight(Name);
        Image->Canvas->Brush->Style = bsClear;
        if (4*Rad > 3*Wt) Image->Canvas->TextOut(Xp - Wt/2, Yp - Ht/2, Name);
        else Image->Canvas->TextOut(Xp + Rad + 10, Yp - Ht/2, Name);
    }
};

// ----------------------------------------------------------------------------
//  :

void TPlanet::DrawTrajectory(CAMERA *Camera)
{
	if (Image == NULL || !HasTrajectory) return;
	int W = Image->Width;
    int H = Image->Height;
    //   :
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double F = double(W)/2/tan(Camera->Phi);		//  
    double **R = Camera->R;							//  
    double MinInt = 1.0/(MAXINT - 4092);

    int PointCount = Trajectory->GetPoints();
    if (PointCount == 0) return;
    TPoint *DataPOINTS = new TPoint[PointCount];
    T3DPoint *Point = Trajectory->Last;
    Image->Canvas->Pen->Color = TrajColor;
    Image->Canvas->Pen->Mode = pmNotXor;
    double XR,YR,ZR,C;
    int Xp,Yp;
    int k = 0;
    bool Redraw = false;
    while (Point!=NULL) {
    	XR = R[0][0]*(Point->X-X0) + R[0][1]*(Point->Y-Y0) + R[0][2]*(Point->Z-Z0);
    	YR = R[1][0]*(Point->X-X0) + R[1][1]*(Point->Y-Y0) + R[1][2]*(Point->Z-Z0);
    	ZR = R[2][0]*(Point->X-X0) + R[2][1]*(Point->Y-Y0) + R[2][2]*(Point->Z-Z0);
        C = -ZR/F;
    	if (!(C > MinInt*((XR>0)?XR:-XR) && C > MinInt*((YR>0)?YR:-YR) && C > 0)) Redraw = true;
        else {
    		Xp = W/2 + trunc(XR/C);
    		Yp = H/2 - trunc(YR/C);
            if (!(-4000<Xp && Xp<4000 && -4000<Yp && Yp<4000)) Redraw = true;
        }
        if (!Redraw) {
        	DataPOINTS[k].x = Xp;
            DataPOINTS[k].y = Yp;
            k++;
        }
        else {
        	if (k>0) {
        		Image->Canvas->Polyline(DataPOINTS, k-1);
            	k = 0;
        	}
            Redraw = false;
        }
        Point = Point->Prev;
    }
    Image->Canvas->Polyline(DataPOINTS, k-1);
    delete[] DataPOINTS;
};

